# -*- coding: utf-8 -*-
"""Samila errors."""


class samilaDataError(Exception):
    """Data error class."""

    pass


class samilaGenerateError(Exception):
    """Generate error class."""

    pass


class samilaPlotError(Exception):
    """Plot error class."""

    pass


class samilaConfigError(Exception):
    """Config error class."""

    pass
